# -*- coding: utf-8 -*-

"""
Implementation for the ``Issho`` class, which implements
a connection and some simple commands over ``ssh``, using
``keyring`` to manage secrets locally.
"""

import paramiko
import keyring
from sshtunnel import SSHTunnelForwarder
from issho.helpers import default_sftp_path, get_pkey, issho_pw_name, get_user
from issho.config import read_issho_conf, read_ssh_profile
import sys
import time
from shutil import copyfile
import humanize


class Issho:
    def __init__(self, profile="dev", kinit=True):
        self.local_user = get_user()
        self.profile = profile
        self.issho_conf = read_issho_conf(profile)
        self.ssh_conf = read_ssh_profile(self.issho_conf["SSH_CONFIG_PATH"], profile)
        self.hostname = self.ssh_conf.get("hostname", None)
        self.user = self.ssh_conf.get("user", None)
        self.port = self.ssh_conf.get("port", 22)
        self._ssh = self._connect()
        if kinit:
            self.kinit()
        self._remote_home_dir = self.get_output("echo $HOME").strip()
        return

    def local_forward(
        self, remote_host, remote_port, local_host="0.0.0.0", local_port=44556
    ):
        """
        Forwards a port from a remote through this Issho object.
        Useful for connecting to remote hosts that can only be accessed
        from inside a VPC of which your devbox is part.
        """
        tunnel = SSHTunnelForwarder(
            (self.hostname, self.port),
            ssh_username=self.user,
            ssh_pkey=get_pkey(self.issho_conf["ID_RSA"]),
            remote_bind_address=(remote_host, remote_port),
            local_bind_address=(local_host, local_port),
        )
        tunnel.start()
        return tunnel

    def exec(self, cmd, bg=False, debug=False, capture_output=False):
        """
        Execute a command in bash over the SSH connection.

        Note, this command does not use an interactive terminal;
        it instead uses a *non-interactive login* shell.
        This means (specifically) that your aliased commands will not work
        and only variables exported in your remote ``.bashrc`` will be available.

        :param cmd: The bash command to be run remotely

        :param bg: True = run in the background

        :param debug: True = print some debugging output

        :param capture_output: True = return stdout as a string

        :return:
        """
        if bg:
            cmd = 'cmd=$"{}"; nohup bash -c "$cmd" &'.format(cmd.replace('"', r"\""))
        if debug:
            print(cmd)
        stdin, stdout, stderr = self._ssh.exec_command(cmd)

        captured_output = ""
        for line in stdout:
            if capture_output:
                captured_output += line
            else:
                print(line, end="")

        for line in stderr:
            sys.stderr.write(line)
        return captured_output

    def exec_bg(self, cmd, **kwargs):
        """
        Syntactic sugar for ``exec(bg=True)``
        """
        return self.exec(cmd, bg=True, **kwargs)

    def get_output(self, cmd, **kwargs):
        """
        Syntactic sugar for ``exec(capture_output=True)``
        """
        return self.exec(cmd, capture_output=True, **kwargs)

    def get(self, remotepath, localpath=None):
        """
        Gets the file at the remote path and puts it locally.

        :param remotepath: The path on the remote from which to get.

        :param localpath: Defaults to the name of the remote path
        """
        paths = self._sftp_paths(localpath=localpath, remotepath=remotepath)
        with self._ssh.open_sftp() as sftp:
            sftp.get(
                remotepath=paths["remotepath"],
                localpath=paths["localpath"],
                callback=self._sftp_progress,
            )
        return

    def put(self, localpath, remotepath=None):
        """
        Puts the file at the local path to the remote.

        :param localpath: The local path of the file to put to the remote

        :param remotepath: Defaults to the name of the local path
        """
        paths = self._sftp_paths(localpath=localpath, remotepath=remotepath)
        with self._ssh.open_sftp() as sftp:
            sftp.put(
                localpath=paths["localpath"],
                remotepath=paths["remotepath"],
                callback=self._sftp_progress,
            )
        return

    def kinit(self):
        """
        Runs kerberos init
        """
        kinit_pw = self._get_password("kinit")
        if kinit_pw:
            self.exec("echo {} | kinit".format(kinit_pw))
        else:
            raise OSError(
                "Add your kinit password with `issho config <profile>` "
                "or by editing `~/.issho/config.toml`"
            )
        return

    def hive(self, query, output_filename=None, remove_blank_top_line=True):
        """
        Runs a hive query using the parameters
        set in .issho/config.toml

        :param query: a string query, or the name of a query file
            name to run.
        :param output_filename: the (local) file to output the results
            of the hive query to. Adding this option will also
            keep a copy of the results in /tmp
        :param remove_blank_top_line: Hive usually has a blank top line
            when data is output, this parameter removes it.
        """
        query = str(query)
        tmp_filename = "/tmp/issho_{}.sql".format(time.time())
        if query.endswith("sql") or query.endswith("hql"):
            copyfile(query, tmp_filename)
        else:
            with open(tmp_filename, "w") as f:
                f.write(query)
        self.put(tmp_filename, tmp_filename)

        tmp_output_filename = "{}.output".format(tmp_filename)

        hive_cmd = 'beeline {opts} -u  "{jdbc}" -f {fn} {remove_first_line} {redirect_to_tmp_fn}'.format(
            opts=self.issho_conf["HIVE_OPTS"],
            jdbc=self.issho_conf["HIVE_JDBC"],
            fn=tmp_filename,
            remove_first_line="| sed 1d" if remove_blank_top_line else "",
            redirect_to_tmp_fn="> {}".format(tmp_output_filename)
            if output_filename
            else "",
        )

        self.exec(hive_cmd)

        if output_filename:
            self.get(tmp_output_filename, output_filename)

    def _connect(self):
        """
        Uses paramiko to connect to the remote specified
        :return:
        """
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(
            self.hostname,
            username=self.user,
            port=self.port,
            pkey=get_pkey(self.issho_conf["RSA_ID_PATH"]),
        )
        return ssh

    def _sftp_paths(self, localpath, remotepath):
        localpath = default_sftp_path(localpath, remotepath)
        remotepath = default_sftp_path(remotepath, localpath)
        return {
            "localpath": str(localpath.expanduser()),
            "remotepath": str(remotepath).replace("~", self._remote_home_dir),
        }

    @staticmethod
    def _sftp_progress(transferred, to_transfer):
        print(
            "{} transferred out of a total of {}".format(
                humanize.naturalsize(transferred), humanize.naturalsize(to_transfer)
            )
        )

    def _get_password(self, pw_type):
        return keyring.get_password(
            issho_pw_name(pw_type=pw_type, profile=self.profile), self.local_user
        )



from prompt_toolkit import prompt
import keyring
from issho.config import (
    write_issho_conf,
    read_issho_conf,
    read_ssh_profile,
    write_issho_env,
    read_issho_env,
)
from issho.helpers import issho_pw_name, issho_ssh_pw_name, absolute_path, get_user
from collections import OrderedDict
from issho.issho import Issho
import fire
import re

OPENSSH_PASSWORD_ERROR = """
Paramiko v2.4.0 does not allow OpenSSH RSA key format (common on new Macs);
see: https://github.com/paramiko/paramiko/issues/1313#issuecomment-492448807
Create your ssh key using:
$ ssh-keygen -t rsa -b 4096 -C "email@email.com" -m PEM
"""

ENV_PROMPTS = OrderedDict(
    (
        ("HIVE_OPTS", "Hive Options: "),
        ("HIVE_JDBC", "Hive JDBC connection string: "),
        ("SPARK_CONF", "Spark Shell Configuration String: "),
    )
)


class IsshoCLI:
    """
    CLI for Issho; right now only used for configuration
    """

    def config(
        self,
        profile,
        env=None,
        ssh_profile="",
        ssh_config="~/.ssh/config",
        rsa_id="~/.ssh/id_rsa",
    ):
        """
        Configures a single issho profile. Saves non-private variables
        to ``~/.issho/conf.toml`` and passwords to the local keyring.

        :param profile: name of the profile to configure

        :param env: Optional environment variable profile to draw from.

        :param ssh_profile: The name of the associated ssh config profile;
            defaults to the profile name if not supplied.

        :param ssh_config: the path to the ssh_config to be used for this profile

        :param rsa_id: the path to the id_rsa file to be used for this profile
        """

        local_user = get_user()
        ssh_profile = profile if not ssh_profile else profile
        rsa_id = absolute_path(rsa_id)
        ssh_config = absolute_path(ssh_config)
        env = read_issho_env(env) if env else {}

        ssh_conf = read_ssh_profile(ssh_config, ssh_profile)
        if not all(x in ssh_conf for x in ("hostname", "user")):
            raise KeyError()

        if not keyring.get_password(issho_ssh_pw_name(rsa_id), rsa_id):
            _set_up_ssh_password(rsa_id=rsa_id)

        kinit_was_setup = _set_up_password(
            pw_type="kinit", profile=profile, pw_user=local_user
        )

        new_conf = {
            "SSH_CONFIG_PATH": ssh_config,
            "RSA_ID_PATH": rsa_id,
            **_get_env_vars(env),
        }
        write_issho_conf({profile: new_conf})
        self.test_connection(profile, kinit=kinit_was_setup)

    @staticmethod
    def env(env_name):
        """
        Saves a set of variables to ~/.issho/envs.toml
        :param env_name: name of the environment to set up or update
        """
        env_conf = {env_name: _get_env_vars({})}
        write_issho_env(env_conf)
        return env_conf

    @staticmethod
    def env(env_name):
        """
        Saves a set of variables to ~/.issho/envs.toml
        :param env_name: name of the environment to set up or update
        """
        env_conf = {env_name: _get_env_vars({})}
        write_issho_env(env_conf)
        return env_conf

    @staticmethod
    def update_variable(profile, variable, value):
        """
        Updates or add a single profile variable.
        """
        old_conf = read_issho_conf(profile)
        old_conf[variable] = value
        write_issho_conf({profile: old_conf})

    @staticmethod
    def test_connection(profile, kinit=True):
        """
        Tests the connection to the specified
        :param profile: The name of the profile
        :param kinit: if True, will try to kinit using the stored password
        :return:
        """
        try:
            test_issho = Issho(profile, kinit)
        except Exception as e:
            return "Test Connection failed with error: {}".format(str(e))
        return "Test Connection Successful!"


def _get_env_vars(env):
    """
    Gets the set of parameters described in ENV_PROMPTS
    using a pre-existing environment env if it exists

    :param env: a dictionary containing previously-set environment variables

    :return:
    """
    return {
        var_name: env.get(var_name) if var_name in env else prompt(prompt_str)
        for var_name, prompt_str in ENV_PROMPTS.items()
    }


def _get_pw(pw_type):
    """
    Gets a password using the prompt toolkit
    """
    while True:
        pw = prompt("Enter the {} password: ".format(pw_type), is_password=True)
        if not pw:
            break
        pw2 = prompt("Enter the {} password again: ".format(pw_type), is_password=True)
        if pw != pw2:
            print("The passwords do not match; try again")
        else:
            return pw


def _keep_old_password(pw_type, pw_name, pw_user):
    """
    :return True to keep the old password, False to reset
    """
    if keyring.get_password(pw_name, pw_user):
        new_pw_prompt = (
            "A {} password exists for {}--do you want to set a new password (y/N)? "
        )
        reset_password = prompt(new_pw_prompt.format(pw_type, pw_user))
        return (not reset_password) or (
            not reset_password.strip().lower().startswith("y")
        )
    return False


def _set_up_password(pw_type, profile, pw_user):
    """
    Gets an issho password for this profile and
    saves it to the local keyring.
    """
    pw_name = issho_pw_name(pw_type=pw_type, profile=profile)
    if _keep_old_password(pw_type, pw_name, pw_user):
        return True if keyring.get_password(pw_name, pw_user) else False
    else:
        pw = _get_pw(pw_type=pw_type)
        keyring.set_password(pw_name, pw_user, pw)
    return True if pw else False


def _set_up_ssh_password(rsa_id):
    """
    Adds the ssh password to the local keyring.
    """
    _check_not_openssh_pkey(rsa_id)
    pw = _get_pw(pw_type="ssh")
    pw_name = issho_ssh_pw_name(rsa_id=rsa_id)
    keyring.set_password(pw_name, rsa_id, pw)
    return True if pw else False


def _check_not_openssh_pkey(rsa_id):
    with open(rsa_id) as f:
        if re.search("OPENSSH", f.readline()):
            raise ValueError(OPENSSH_PASSWORD_ERROR)


def main():
    """
    Inititates the CLI using python-fire
    """
    fire.Fire(IsshoCLI)


if __name__ == "__main__":
    main()



from pathlib import Path
import socket
import paramiko
import keyring
import os


def absolute_path(raw_path):
    """
    Gets the string absolute path from a path object or string.

    :param raw_path: a string or ``pathlib.Path`` object
    """
    if not isinstance(raw_path, Path):
        raw_path = Path(raw_path)
    return str(raw_path.expanduser())


def default_sftp_path(this_path, default_path):
    """
    If ``this_path`` exists, return it as a path, else, return
    the ``pathlib.Path.name`` of the default path
    """
    return Path(this_path) if this_path else Path(Path(default_path).name)


def able_to_connect(host, port, timeout=1.5):
    """
    Returns true if it is possible to connect to the specified host
    and port, within the given timeout in seconds.
    """
    try:
        sock = socket.socket()
        sock.settimeout(timeout)
        sock.connect((host, port))
    except Exception as e:
        return False
    return True


def get_pkey(key_path):
    """
    Helper for getting an RSA key
    """
    key_file = absolute_path(key_path)
    return paramiko.RSAKey.from_private_key_file(
        key_file, password=keyring.get_password(issho_ssh_pw_name(key_file), key_file)
    )


def issho_pw_name(pw_type, profile):
    """
    Helper for standardizing password names
    """
    return "issho_{}_{}".format(pw_type, profile)


def issho_ssh_pw_name(rsa_id):
    """
    Helper for standardizing ssh password names
    """
    return "issho_ssh_{}".format(
        "".join(ch for ch in absolute_path(rsa_id) if ch.isalnum())
    )


def get_user():
    return os.environ.get("USER")



import toml
import paramiko
from pathlib import Path
from issho.helpers import absolute_path

ISSHO_DIR = Path.home().joinpath(".issho")
ISSHO_CONF_FILE = ISSHO_DIR.joinpath("conf.toml")
ISSHO_ENV_FILE = ISSHO_DIR.joinpath("envs.toml")


def _make_issho_conf_dir():
    if not ISSHO_DIR.exists():
        ISSHO_DIR.mkdir()
    if not ISSHO_CONF_FILE.exists():
        ISSHO_CONF_FILE.touch()
    if not ISSHO_ENV_FILE.exists():
        ISSHO_ENV_FILE.touch()
    return


def read_issho_conf(profile, filename=ISSHO_CONF_FILE):
    """
    Writes issho variables out to a ``.toml`` file.

    :param profile: The name of the profile to read
    :param filename: The output filename
    :return: a dict of data stored with that profile in the configuration file
    """
    _make_issho_conf_dir()
    conf = toml.load(filename)
    if profile not in conf:
        raise ValueError
    return conf[profile]


def read_issho_env(profile):
    """
    Reads issho environment variables to a dict
    :param profile: the name of the issho environment to draw from
    :return: a dict of data with that profile stored in the environment file
    """
    return read_issho_conf(profile, filename=ISSHO_ENV_FILE)


def write_issho_conf(new_conf_dict, filename=ISSHO_CONF_FILE):
    """
    Updates the issho config file
    :param new_conf_dict: the new configuration to add
    :param filename: the location of the old configuration file
    """
    _make_issho_conf_dir()
    old_issho_conf = toml.load(filename)
    new_conf = {**old_issho_conf, **new_conf_dict}
    toml.dump(new_conf, open(str(filename), "w"))
    return


def write_issho_env(new_env_dict):
    """
    Save a new issho environment
    :param new_env_dict: the new set of environment paramters to add
    """
    write_issho_conf(new_env_dict, filename=ISSHO_ENV_FILE)


def read_ssh_profile(ssh_config_path, profile):
    """
    Helper method for getting data from .ssh/config
    """
    ssh_config_file = absolute_path(ssh_config_path)
    conf = paramiko.SSHConfig()
    conf.parse(open(ssh_config_file))
    return conf.lookup(profile)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "sshtunnel>=0.1.4",
    "keyring>=18.0.0",
    "paramiko>=2.4.0",
    "cryptography==2.4.2",  # https://github.com/paramiko/paramiko/issues/1379
    # https://github.com/paramiko/paramiko/issues/1446
    "prompt_toolkit>=1.0.10",  # prompt_toolkit is used by jupyter_console,
    "toml>=0.10.0",
    "fire>=0.1.3",
    "humanize>=0.5.0",
]

setup_requirements = []

test_requirements = []

setup(
    author="Michael Bilow",
    author_email="michael.k.bilow@gmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="issho allows easy access to simple commands on a remote machine.",
    entry_points="""
        [console_scripts]
        issho=issho.cli:main
    """,
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="issho,paramiko,keyring,ssh,sftp,python,python-fire,prompt-toolkit,pysftp,sshed,sshtunnel",
    name="issho",
    packages=find_packages(include=["issho"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/michaelbilow/issho",
    version="0.4.1",
    zip_safe=False,
)
